# distutils: language=c++
# distutils: sources=hummingbot/core/cpp/PyRef.cpp

from cpython cimport(
    PyObject,
    PyWeakref_NewRef,
    PyWeakref_GetObject
)
from cython.operator cimport(
    postincrement as inc,
    dereference as deref,
    address
)
from libcpp.vector cimport vector
from enum import Enum
import logging
import random
from typing import List

from hummingbot.logger import HummingbotLogger
from hummingbot.core.event.event_listener import EventListener
from hummingbot.core.event.event_listener cimport EventListener

class_logger = None


cdef class PubSub:
    """
    PubSub with weak references. This avoids the lapsed listener problem by periodically performing GC on dead
    event listener.

    Dead listener is done by calling c_remove_dead_listeners(), which checks whether the listener weak references are
    alive or not, and removes the dead ones. Each call to c_remove_dead_listeners() takes O(n).

    Here's how the dead listener GC is performed:

    1. c_add_listener():
       Randomly with ADD_LISTENER_GC_PROBABILITY. This assumes c_add_listener() is called frequently and so it doesn't
       make sense to do the GC every time.
    2. c_remove_listener():
       Every time. This assumes c_remove_listener() is called infrequently.
    3. c_get_listeners() and c_trigger_event():
       Every time. Both functions take O(n) already.
    """

    ADD_LISTENER_GC_PROBABILITY = 0.005

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global class_logger
        if class_logger is None:
            class_logger = logging.getLogger(__name__)
        return class_logger

    def __init__(self):
        self._events = Events()

    def add_listener(self, event_tag: Enum, listener: EventListener):
        self.c_add_listener(event_tag.value, listener)

    def remove_listener(self, event_tag: Enum, listener: EventListener):
        self.c_remove_listener(event_tag.value, listener)

    def get_listeners(self, event_tag: Enum) -> List[EventListener]:
        return self.c_get_listeners(event_tag.value)

    def trigger_event(self, event_tag: Enum, message: any):
        self.c_trigger_event(event_tag.value, message)

    cdef c_log_exception(self, int64_t event_tag, object arg):
        self.logger().error(f"Unexpected error while processing event {event_tag}.", exc_info=True)

    cdef c_add_listener(self, int64_t event_tag, EventListener listener):
        cdef:
            EventsIterator it = self._events.find(event_tag)
            EventListenersCollection new_listeners
            EventListenersCollection *listeners_ptr
            object listener_weakref = PyWeakref_NewRef(listener, None)
            PyRef listener_wrapper = PyRef(<PyObject *>listener_weakref)
        if it != self._events.end():
            listeners_ptr = address(deref(it).second)
            deref(listeners_ptr).insert(listener_wrapper)
        else:
            new_listeners.insert(listener_wrapper)
            self._events.insert(EventsPair(event_tag, new_listeners))

        if random.random() < PubSub.ADD_LISTENER_GC_PROBABILITY:
            self.c_remove_dead_listeners(event_tag)

    cdef c_remove_listener(self, int64_t event_tag, EventListener listener):
        cdef:
            EventsIterator it = self._events.find(event_tag)
            EventListenersCollection *listeners_ptr
            object listener_weakref = PyWeakref_NewRef(listener, None)
            PyRef listener_wrapper = PyRef(<PyObject *>listener_weakref)
            EventListenersIterator lit
        if it == self._events.end():
            return
        listeners_ptr = address(deref(it).second)
        lit = deref(listeners_ptr).find(listener_wrapper)
        if lit != deref(listeners_ptr).end():
            deref(listeners_ptr).erase(lit)
        self.c_remove_dead_listeners(event_tag)

    cdef c_remove_dead_listeners(self, int64_t event_tag):
        cdef:
            EventsIterator it = self._events.find(event_tag)
            EventListenersCollection *listeners_ptr
            object listener_weakref
            EventListenersIterator lit
            vector[EventListenersIterator] lit_to_remove
        if it == self._events.end():
            return
        listeners_ptr = address(deref(it).second)
        lit = deref(listeners_ptr).begin()
        while lit != deref(listeners_ptr).end():
            listener_weakref = <object>(deref(lit).get())
            if <object>(PyWeakref_GetObject(listener_weakref)) is None:
                lit_to_remove.push_back(lit)
            inc(lit)
        for lit in lit_to_remove:
            deref(listeners_ptr).erase(lit)
        if deref(listeners_ptr).size() < 1:
            self._events.erase(it)

    cdef c_get_listeners(self, int64_t event_tag):
        self.c_remove_dead_listeners(event_tag)

        cdef:
            EventsIterator it = self._events.find(event_tag)
            EventListenersCollection *listeners_ptr
            object listener_weafref
            EventListener typed_listener

        if it == self._events.end():
            return []

        retval = []
        listeners_ptr = address(deref(it).second)
        for pyref in deref(listeners_ptr):
            listener_weafref = <object>pyref.get()
            typed_listener = <object>PyWeakref_GetObject(listener_weafref)
            retval.append(typed_listener)
        return retval

    cdef c_trigger_event(self, int64_t event_tag, object arg):
        self.c_remove_dead_listeners(event_tag)

        cdef:
            EventsIterator it = self._events.find(event_tag)
            EventListenersCollection listeners
            object listener_weafref
            EventListener typed_listener
        if it == self._events.end():
            return

        # It is extremely important that this set of listeners is a C++ copy - because listeners are allowed to call
        # c_remove_listener(), which breaks the iterator if we're using the underlying set.
        listeners = deref(it).second
        for pyref in listeners:
            listener_weafref = <object>pyref.get()
            typed_listener = <object>PyWeakref_GetObject(listener_weafref)
            try:
                typed_listener.c_set_event_info(event_tag, self)
                typed_listener.c_call(arg)
            except Exception:
                self.c_log_exception(event_tag, arg)
            finally:
                typed_listener.c_set_event_info(0, None)



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

DEFAULT_DOMAIN = "com"

HBOT_ORDER_ID_PREFIX = "HUMBOT"
MAX_ORDER_ID_LEN = 32

# Base URL
REST_URL = "https://api.mexc.{}/api/"
WSS_URL = "wss://wbs-api.mexc.{}/ws"

PUBLIC_API_VERSION = "v3"
PRIVATE_API_VERSION = "v3"

# Public API endpoints or MexcClient function
TICKER_PRICE_CHANGE_PATH_URL = "/ticker/24hr"
TICKER_BOOK_PATH_URL = "/ticker/bookTicker"
EXCHANGE_INFO_PATH_URL = "/exchangeInfo"
SUPPORTED_SYMBOL_PATH_URL = "/defaultSymbols"
PING_PATH_URL = "/ping"
SNAPSHOT_PATH_URL = "/depth"
SERVER_TIME_PATH_URL = "/time"

# Private API endpoints or MexcClient function
ACCOUNTS_PATH_URL = "/account"
MY_TRADES_PATH_URL = "/myTrades"
ORDER_PATH_URL = "/order"
MEXC_USER_STREAM_PATH_URL = "/userDataStream"

WS_HEARTBEAT_TIME_INTERVAL = 30

# Mexc params

SIDE_BUY = "BUY"
SIDE_SELL = "SELL"

TIME_IN_FORCE_GTC = "GTC"  # Good till cancelled
TIME_IN_FORCE_IOC = "IOC"  # Immediate or cancel
TIME_IN_FORCE_FOK = "FOK"  # Fill or kill

# Rate Limit Type
IP_REQUEST_WEIGHT = "IP_REQUEST_WEIGHT"
UID_REQUEST_WEIGHT = "UID_REQUEST_WEIGHT"

# Rate Limit time intervals
ONE_MINUTE = 60
ONE_SECOND = 1
ONE_DAY = 86400

MAX_REQUEST = 5000

# Order States
ORDER_STATE = {
    "PENDING": OrderState.PENDING_CREATE,
    "NEW": OrderState.OPEN,
    "FILLED": OrderState.FILLED,
    "PARTIALLY_FILLED": OrderState.PARTIALLY_FILLED,
    "PENDING_CANCEL": OrderState.OPEN,
    "PARTIALLY_CANCELED": OrderState.CANCELED,
    "CANCELED": OrderState.CANCELED,
    "REJECTED": OrderState.FAILED,
    "EXPIRED": OrderState.FAILED,
}

# WS Order States
WS_ORDER_STATE = {
    1: OrderState.OPEN,
    2: OrderState.FILLED,
    3: OrderState.PARTIALLY_FILLED,
    4: OrderState.CANCELED,
    5: OrderState.OPEN,
}

# Websocket event types
PUBLIC_TRADES_ENDPOINT_NAME = "spot@public.aggre.deals.v3.api.pb"
PUBLIC_DIFF_ENDPOINT_NAME = "spot@public.aggre.depth.v3.api.pb"

TRADE_EVENT_TYPE = "public.aggre.deals"
DIFF_EVENT_TYPE = "public.aggre.depth"

USER_TRADES_ENDPOINT_NAME = "spot@private.deals.v3.api.pb"
USER_ORDERS_ENDPOINT_NAME = "spot@private.orders.v3.api.pb"
USER_BALANCE_ENDPOINT_NAME = "spot@private.account.v3.api.pb"
WS_CONNECTION_TIME_INTERVAL = 20
RATE_LIMITS = [
    RateLimit(limit_id=IP_REQUEST_WEIGHT, limit=20000, time_interval=ONE_MINUTE),
    RateLimit(limit_id=UID_REQUEST_WEIGHT, limit=240000, time_interval=ONE_MINUTE),
    # Weighted Limits
    RateLimit(limit_id=TICKER_PRICE_CHANGE_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 1)]),
    RateLimit(limit_id=TICKER_BOOK_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 2)]),
    RateLimit(limit_id=EXCHANGE_INFO_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 10)]),
    RateLimit(limit_id=SUPPORTED_SYMBOL_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 10)]),
    RateLimit(limit_id=SNAPSHOT_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 50)]),
    RateLimit(limit_id=MEXC_USER_STREAM_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(UID_REQUEST_WEIGHT, 1)]),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 1)]),
    RateLimit(limit_id=PING_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(IP_REQUEST_WEIGHT, 1)]),
    RateLimit(limit_id=ACCOUNTS_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(UID_REQUEST_WEIGHT, 10)]),
    RateLimit(limit_id=MY_TRADES_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(UID_REQUEST_WEIGHT, 10)]),
    RateLimit(limit_id=ORDER_PATH_URL, limit=MAX_REQUEST, time_interval=ONE_MINUTE,
              linked_limits=[LinkedLimitWeightPair(UID_REQUEST_WEIGHT, 2)])
]

ORDER_NOT_EXIST_ERROR_CODE = -2013
ORDER_NOT_EXIST_MESSAGE = "Order does not exist"
UNKNOWN_ORDER_ERROR_CODE = -2011
UNKNOWN_ORDER_MESSAGE = "Unknown order sent"
TIMESTAMP_RELATED_ERROR_CODE = 700003
TIMESTAMP_RELATED_ERROR_MESSAGE = "Timestamp for this request is outside of the recvWindow"



from typing import TYPE_CHECKING

import pandas as pd

from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.core.utils.async_utils import safe_ensure_future

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401

import threading


class OrderBookCommand:
    def order_book(self,  # type: HummingbotApplication
                   lines: int = 5,
                   exchange: str = None,
                   market: str = None,
                   live: bool = False):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.order_book, lines, exchange, market, live)
            return
        safe_ensure_future(self.show_order_book(lines, exchange, market, live))

    async def show_order_book(self,  # type: HummingbotApplication
                              lines: int = 5,
                              exchange: str = None,
                              market: str = None,
                              live: bool = False):
        if len(self.trading_core.markets.keys()) == 0:
            self.notify("There is currently no active market.")
            return
        if exchange is not None:
            if exchange not in self.trading_core.markets:
                self.notify("Invalid exchange")
                return
            market_connector = self.trading_core.markets[exchange]
        else:
            market_connector = list(self.trading_core.markets.values())[0]
        if market is not None:
            market = market.upper()
            if market not in market_connector.order_books:
                self.notify("Invalid market")
                return
            trading_pair, order_book = market, market_connector.order_books[market]
        else:
            trading_pair, order_book = next(iter(market_connector.order_books.items()))

        def get_order_book(lines):
            bids = order_book.snapshot[0][['price', 'amount']].head(lines)
            bids.rename(columns={'price': 'bid_price', 'amount': 'bid_volume'}, inplace=True)
            asks = order_book.snapshot[1][['price', 'amount']].head(lines)
            asks.rename(columns={'price': 'ask_price', 'amount': 'ask_volume'}, inplace=True)
            joined_df = pd.concat([bids, asks], axis=1)
            text_lines = [
                "    " + line
                for line in format_df_for_printout(joined_df, self.client_config_map.tables_format).split("\n")
            ]
            header = f"  market: {market_connector.name} {trading_pair}\n"
            return header + "\n".join(text_lines)

        if live:
            await self.stop_live_update()
            self.app.live_updates = True
            while self.app.live_updates:
                await self.cls_display_delay(get_order_book(min(lines, 35)) + "\n\n Press escape key to stop update.", 0.5)
            self.notify("Stopped live orderbook display update.")
        else:
            self.notify(get_order_book(lines))



from decimal import Decimal
from typing import List, NamedTuple

from hummingbot.core.data_type.common import OrderType

ORDER_PROPOSAL_ACTION_CREATE_ORDERS = 1
ORDER_PROPOSAL_ACTION_CANCEL_ORDERS = 1 << 1


class OrdersProposal(NamedTuple):
    actions: int
    buy_order_type: OrderType
    buy_order_prices: List[Decimal]
    buy_order_sizes: List[Decimal]
    sell_order_type: OrderType
    sell_order_prices: List[Decimal]
    sell_order_sizes: List[Decimal]
    cancel_order_ids: List[str]


class PricingProposal(NamedTuple):
    buy_order_prices: List[Decimal]
    sell_order_prices: List[Decimal]


class SizingProposal(NamedTuple):
    buy_order_sizes: List[Decimal]
    sell_order_sizes: List[Decimal]


class InventorySkewBidAskRatios(NamedTuple):
    bid_ratio: float
    ask_ratio: float


class PriceSize:
    def __init__(self, price: Decimal, size: Decimal):
        self.price: Decimal = price
        self.size: Decimal = size

    def __repr__(self):
        return f"[ p: {self.price} s: {self.size} ]"


class Proposal:
    def __init__(self, buys: List[PriceSize], sells: List[PriceSize]):
        self.buys: List[PriceSize] = buys
        self.sells: List[PriceSize] = sells

    def __repr__(self):
        return f"{len(self.buys)} buys: {', '.join([str(o) for o in self.buys])} " \
               f"{len(self.sells)} sells: {', '.join([str(o) for o in self.sells])}"



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.binance import (
    binance_constants as CONSTANTS,
    binance_utils,
    binance_web_utils as web_utils,
)
from hummingbot.connector.exchange.binance.binance_api_order_book_data_source import BinanceAPIOrderBookDataSource
from hummingbot.connector.exchange.binance.binance_api_user_stream_data_source import BinanceAPIUserStreamDataSource
from hummingbot.connector.exchange.binance.binance_auth import BinanceAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import TradeFillOrderDetails, combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import DeductedFromReturnsTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.event.events import MarketEvent, OrderFilledEvent
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class BinanceExchange(ExchangePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    def __init__(self,
                 binance_api_key: str,
                 binance_api_secret: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 ):
        self.api_key = binance_api_key
        self.secret_key = binance_api_secret
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trades_poll_binance_timestamp = 1.0
        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @staticmethod
    def binance_order_type(order_type: OrderType) -> str:
        return order_type.name.upper()

    @staticmethod
    def to_hb_order_type(binance_type: str) -> OrderType:
        return OrderType[binance_type]

    @property
    def authenticator(self):
        return BinanceAuth(
            api_key=self.api_key,
            secret_key=self.secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        if self._domain == "com":
            return "binance"
        else:
            return f"binance_{self._domain}"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.PING_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def get_all_pairs_prices(self) -> List[Dict[str, str]]:
        pairs_prices = await self._api_get(path_url=CONSTANTS.TICKER_BOOK_PATH_URL)
        return pairs_prices

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        is_time_synchronizer_related = ("-1021" in error_description
                                        and "Timestamp for this request" in error_description)
        return is_time_synchronizer_related

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return str(CONSTANTS.ORDER_NOT_EXIST_ERROR_CODE) in str(
            status_update_exception
        ) and CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return str(CONSTANTS.UNKNOWN_ORDER_ERROR_CODE) in str(
            cancelation_exception
        ) and CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BinanceAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            domain=self.domain,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BinanceAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = order_type is OrderType.LIMIT_MAKER
        return DeductedFromReturnsTradeFee(percent=self.estimate_fee_pct(is_maker))

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        order_result = None
        amount_str = f"{amount:f}"
        type_str = BinanceExchange.binance_order_type(order_type)
        side_str = CONSTANTS.SIDE_BUY if trade_type is TradeType.BUY else CONSTANTS.SIDE_SELL
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        api_params = {"symbol": symbol,
                      "side": side_str,
                      "quantity": amount_str,
                      "type": type_str,
                      "newClientOrderId": order_id}
        if order_type is OrderType.LIMIT or order_type is OrderType.LIMIT_MAKER:
            price_str = f"{price:f}"
            api_params["price"] = price_str
        if order_type == OrderType.LIMIT:
            api_params["timeInForce"] = CONSTANTS.TIME_IN_FORCE_GTC

        try:
            order_result = await self._api_post(
                path_url=CONSTANTS.ORDER_PATH_URL,
                data=api_params,
                is_auth_required=True)
            o_id = str(order_result["orderId"])
            transact_time = order_result["transactTime"] * 1e-3
        except IOError as e:
            error_description = str(e)
            is_server_overloaded = ("status is 503" in error_description
                                    and "Unknown error, please check your request or try again later." in error_description)
            if is_server_overloaded:
                o_id = "UNKNOWN"
                transact_time = self._time_synchronizer.time()
            else:
                raise
        return o_id, transact_time

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        api_params = {
            "symbol": symbol,
            "origClientOrderId": order_id,
        }
        cancel_result = await self._api_delete(
            path_url=CONSTANTS.ORDER_PATH_URL,
            params=api_params,
            is_auth_required=True)
        if cancel_result.get("status") == "CANCELED":
            return True
        return False

    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        """
        Example:
        {
            "symbol": "ETHBTC",
            "baseAssetPrecision": 8,
            "quotePrecision": 8,
            "orderTypes": ["LIMIT", "MARKET"],
            "filters": [
                {
                    "filterType": "PRICE_FILTER",
                    "minPrice": "0.00000100",
                    "maxPrice": "100000.00000000",
                    "tickSize": "0.00000100"
                }, {
                    "filterType": "LOT_SIZE",
                    "minQty": "0.00100000",
                    "maxQty": "100000.00000000",
                    "stepSize": "0.00100000"
                }, {
                    "filterType": "MIN_NOTIONAL",
                    "minNotional": "0.00100000"
                }
            ]
        }
        """
        trading_pair_rules = exchange_info_dict.get("symbols", [])
        retval = []
        for rule in filter(binance_utils.is_exchange_information_valid, trading_pair_rules):
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule.get("symbol"))
                filters = rule.get("filters")
                price_filter = [f for f in filters if f.get("filterType") == "PRICE_FILTER"][0]
                lot_size_filter = [f for f in filters if f.get("filterType") == "LOT_SIZE"][0]
                min_notional_filter = [f for f in filters if f.get("filterType") in ["MIN_NOTIONAL", "NOTIONAL"]][0]

                min_order_size = Decimal(lot_size_filter.get("minQty"))
                tick_size = price_filter.get("tickSize")
                step_size = Decimal(lot_size_filter.get("stepSize"))
                min_notional = Decimal(min_notional_filter.get("minNotional"))

                retval.append(
                    TradingRule(trading_pair,
                                min_order_size=min_order_size,
                                min_price_increment=Decimal(tick_size),
                                min_base_amount_increment=Decimal(step_size),
                                min_notional_size=Decimal(min_notional)))

            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {rule}. Skipping.")
        return retval

    async def _status_polling_loop_fetch_updates(self):
        await self._update_order_fills_from_trades()
        await super()._status_polling_loop_fetch_updates()

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        async for event_message in self._iter_user_event_queue():
            try:
                event_type = event_message.get("e")
                # Refer to https://developers.binance.com/docs/binance-spot-api-docs/user-data-stream
                # As per the order update section in Binance the ID of the order being canceled is under the "C" key
                if event_type == "executionReport":
                    execution_type = event_message.get("x")
                    if execution_type != "CANCELED":
                        client_order_id = event_message.get("c")
                    else:
                        client_order_id = event_message.get("C")

                    if execution_type == "TRADE":
                        tracked_order = self._order_tracker.all_fillable_orders.get(client_order_id)
                        if tracked_order is not None:
                            fee = TradeFeeBase.new_spot_fee(
                                fee_schema=self.trade_fee_schema(),
                                trade_type=tracked_order.trade_type,
                                percent_token=event_message["N"],
                                flat_fees=[TokenAmount(amount=Decimal(event_message["n"]), token=event_message["N"])]
                            )
                            trade_update = TradeUpdate(
                                trade_id=str(event_message["t"]),
                                client_order_id=client_order_id,
                                exchange_order_id=str(event_message["i"]),
                                trading_pair=tracked_order.trading_pair,
                                fee=fee,
                                fill_base_amount=Decimal(event_message["l"]),
                                fill_quote_amount=Decimal(event_message["l"]) * Decimal(event_message["L"]),
                                fill_price=Decimal(event_message["L"]),
                                fill_timestamp=event_message["T"] * 1e-3,
                            )
                            self._order_tracker.process_trade_update(trade_update)

                    tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
                    if tracked_order is not None:
                        order_update = OrderUpdate(
                            trading_pair=tracked_order.trading_pair,
                            update_timestamp=event_message["E"] * 1e-3,
                            new_state=CONSTANTS.ORDER_STATE[event_message["X"]],
                            client_order_id=client_order_id,
                            exchange_order_id=str(event_message["i"]),
                        )
                        self._order_tracker.process_order_update(order_update=order_update)

                elif event_type == "outboundAccountPosition":
                    balances = event_message["B"]
                    for balance_entry in balances:
                        asset_name = balance_entry["a"]
                        free_balance = Decimal(balance_entry["f"])
                        total_balance = Decimal(balance_entry["f"]) + Decimal(balance_entry["l"])
                        self._account_available_balances[asset_name] = free_balance
                        self._account_balances[asset_name] = total_balance

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    async def _update_order_fills_from_trades(self):
        """
        This is intended to be a backup measure to get filled events with trade ID for orders,
        in case Binance's user stream events are not working.
        NOTE: It is not required to copy this functionality in other connectors.
        This is separated from _update_order_status which only updates the order status without producing filled
        events, since Binance's get order endpoint does not return trade IDs.
        The minimum poll interval for order status is 10 seconds.
        """
        small_interval_last_tick = self._last_poll_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL
        small_interval_current_tick = self.current_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL
        long_interval_last_tick = self._last_poll_timestamp / self.LONG_POLL_INTERVAL
        long_interval_current_tick = self.current_timestamp / self.LONG_POLL_INTERVAL

        if (long_interval_current_tick > long_interval_last_tick
                or (self.in_flight_orders and small_interval_current_tick > small_interval_last_tick)):
            query_time = int(self._last_trades_poll_binance_timestamp * 1e3)
            self._last_trades_poll_binance_timestamp = self._time_synchronizer.time()
            order_by_exchange_id_map = {}
            for order in self._order_tracker.all_fillable_orders.values():
                order_by_exchange_id_map[order.exchange_order_id] = order

            tasks = []
            trading_pairs = self.trading_pairs
            for trading_pair in trading_pairs:
                params = {
                    "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                }
                if self._last_poll_timestamp > 0:
                    params["startTime"] = query_time
                tasks.append(self._api_get(
                    path_url=CONSTANTS.MY_TRADES_PATH_URL,
                    params=params,
                    is_auth_required=True))

            self.logger().debug(f"Polling for order fills of {len(tasks)} trading pairs.")
            results = await safe_gather(*tasks, return_exceptions=True)

            for trades, trading_pair in zip(results, trading_pairs):

                if isinstance(trades, Exception):
                    self.logger().network(
                        f"Error fetching trades update for the order {trading_pair}: {trades}.",
                        app_warning_msg=f"Failed to fetch trade update for {trading_pair}."
                    )
                    continue
                for trade in trades:
                    exchange_order_id = str(trade["orderId"])
                    if exchange_order_id in order_by_exchange_id_map:
                        # This is a fill for a tracked order
                        tracked_order = order_by_exchange_id_map[exchange_order_id]
                        fee = TradeFeeBase.new_spot_fee(
                            fee_schema=self.trade_fee_schema(),
                            trade_type=tracked_order.trade_type,
                            percent_token=trade["commissionAsset"],
                            flat_fees=[TokenAmount(amount=Decimal(trade["commission"]), token=trade["commissionAsset"])]
                        )
                        trade_update = TradeUpdate(
                            trade_id=str(trade["id"]),
                            client_order_id=tracked_order.client_order_id,
                            exchange_order_id=exchange_order_id,
                            trading_pair=trading_pair,
                            fee=fee,
                            fill_base_amount=Decimal(trade["qty"]),
                            fill_quote_amount=Decimal(trade["quoteQty"]),
                            fill_price=Decimal(trade["price"]),
                            fill_timestamp=trade["time"] * 1e-3,
                        )
                        self._order_tracker.process_trade_update(trade_update)
                    elif self.is_confirmed_new_order_filled_event(str(trade["id"]), exchange_order_id, trading_pair):
                        # This is a fill of an order registered in the DB but not tracked any more
                        self._current_trade_fills.add(TradeFillOrderDetails(
                            market=self.display_name,
                            exchange_trade_id=str(trade["id"]),
                            symbol=trading_pair))
                        self.trigger_event(
                            MarketEvent.OrderFilled,
                            OrderFilledEvent(
                                timestamp=float(trade["time"]) * 1e-3,
                                order_id=self._exchange_order_ids.get(str(trade["orderId"]), None),
                                trading_pair=trading_pair,
                                trade_type=TradeType.BUY if trade["isBuyer"] else TradeType.SELL,
                                order_type=OrderType.LIMIT_MAKER if trade["isMaker"] else OrderType.LIMIT,
                                price=Decimal(trade["price"]),
                                amount=Decimal(trade["qty"]),
                                trade_fee=DeductedFromReturnsTradeFee(
                                    flat_fees=[
                                        TokenAmount(
                                            trade["commissionAsset"],
                                            Decimal(trade["commission"])
                                        )
                                    ]
                                ),
                                exchange_trade_id=str(trade["id"])
                            ))
                        self.logger().info(f"Recreating missing trade in TradeFill: {trade}")

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = int(order.exchange_order_id)
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.MY_TRADES_PATH_URL,
                params={
                    "symbol": trading_pair,
                    "orderId": exchange_order_id
                },
                is_auth_required=True,
                limit_id=CONSTANTS.MY_TRADES_PATH_URL)

            for trade in all_fills_response:
                exchange_order_id = str(trade["orderId"])
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    percent_token=trade["commissionAsset"],
                    flat_fees=[TokenAmount(amount=Decimal(trade["commission"]), token=trade["commissionAsset"])]
                )
                trade_update = TradeUpdate(
                    trade_id=str(trade["id"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=exchange_order_id,
                    trading_pair=trading_pair,
                    fee=fee,
                    fill_base_amount=Decimal(trade["qty"]),
                    fill_quote_amount=Decimal(trade["quoteQty"]),
                    fill_price=Decimal(trade["price"]),
                    fill_timestamp=trade["time"] * 1e-3,
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        updated_order_data = await self._api_get(
            path_url=CONSTANTS.ORDER_PATH_URL,
            params={
                "symbol": trading_pair,
                "origClientOrderId": tracked_order.client_order_id},
            is_auth_required=True)

        new_state = CONSTANTS.ORDER_STATE[updated_order_data["status"]]

        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=str(updated_order_data["orderId"]),
            trading_pair=tracked_order.trading_pair,
            update_timestamp=updated_order_data["updateTime"] * 1e-3,
            new_state=new_state,
        )

        return order_update

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        account_info = await self._api_get(
            path_url=CONSTANTS.ACCOUNTS_PATH_URL,
            is_auth_required=True)

        balances = account_info["balances"]
        for balance_entry in balances:
            asset_name = balance_entry["asset"]
            free_balance = Decimal(balance_entry["free"])
            total_balance = Decimal(balance_entry["free"]) + Decimal(balance_entry["locked"])
            self._account_available_balances[asset_name] = free_balance
            self._account_balances[asset_name] = total_balance
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(binance_utils.is_exchange_information_valid, exchange_info["symbols"]):
            mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base=symbol_data["baseAsset"],
                                                                        quote=symbol_data["quoteAsset"])
        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        resp_json = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL,
            params=params
        )

        return float(resp_json["lastPrice"])



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.hyperliquid import (
    hyperliquid_constants as CONSTANTS,
    hyperliquid_web_utils as web_utils,
)
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.hyperliquid.hyperliquid_exchange import HyperliquidExchange


class HyperliquidAPIUserStreamDataSource(UserStreamTrackerDataSource):

    LISTEN_KEY_KEEP_ALIVE_INTERVAL = 1800  # Recommended to Ping/Update listen key to keep connection alive
    HEARTBEAT_TIME_INTERVAL = 30.0
    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            auth: AuthBase,
            trading_pairs: List[str],
            connector: 'HyperliquidExchange',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DOMAIN,
    ):

        super().__init__()
        self._domain = domain
        self._api_factory = api_factory
        self._auth = auth
        self._ws_assistants: List[WSAssistant] = []
        self._connector = connector
        self._current_listen_key = None
        self._listen_for_user_stream_task = None
        self._last_listen_key_ping_ts = None
        self._trading_pairs: List[str] = trading_pairs

        self.token = None

    @property
    def last_recv_time(self) -> float:
        if self._ws_assistant:
            return self._ws_assistant.last_recv_time
        return 0

    async def _get_ws_assistant(self) -> WSAssistant:
        if self._ws_assistant is None:
            self._ws_assistant = await self._api_factory.get_ws_assistant()
        return self._ws_assistant

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange
        """
        ws: WSAssistant = await self._get_ws_assistant()
        url = f"{web_utils.wss_url(self._domain)}"
        await ws.connect(ws_url=url, ping_timeout=self.HEARTBEAT_TIME_INTERVAL)
        safe_ensure_future(self._ping_thread(ws))
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to order events.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        try:
            orders_change_payload = {
                "method": "subscribe",
                "subscription": {
                    "type": "orderUpdates",
                    "user": self._connector.hyperliquid_address,
                }
            }
            subscribe_order_change_request: WSJSONRequest = WSJSONRequest(
                payload=orders_change_payload,
                is_auth_required=True)

            trades_payload = {
                "method": "subscribe",
                "subscription": {
                    "type": "userFills",
                    "user": self._connector.hyperliquid_address,
                }
            }
            subscribe_trades_request: WSJSONRequest = WSJSONRequest(
                payload=trades_payload,
                is_auth_required=True)
            await websocket_assistant.send(subscribe_order_change_request)
            await websocket_assistant.send(subscribe_trades_request)

            self.logger().info("Subscribed to private order and trades changes channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if event_message.get("error") is not None:
            err_msg = event_message.get("error", {}).get("message", event_message.get("error"))
            raise IOError({
                "label": "WSS_ERROR",
                "message": f"Error received via websocket - {err_msg}."
            })
        elif event_message.get("channel") in [
            CONSTANTS.USER_ORDERS_ENDPOINT_NAME,
            CONSTANTS.USEREVENT_ENDPOINT_NAME,
        ]:
            queue.put_nowait(event_message)

    async def _ping_thread(self, websocket_assistant: WSAssistant,):
        try:
            while True:
                ping_request = WSJSONRequest(payload={"method": "ping"})
                await asyncio.sleep(CONSTANTS.HEARTBEAT_TIME_INTERVAL)
                await websocket_assistant.send(ping_request)
        except Exception as e:
            self.logger().debug(f'ping error {e}')

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                await super()._process_websocket_messages(
                    websocket_assistant=websocket_assistant,
                    queue=queue)
            except asyncio.TimeoutError:
                ping_request = WSJSONRequest(payload={"method": "ping"})
                await websocket_assistant.send(ping_request)



from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True
EXAMPLE_PAIR = "ZRX-ETH"

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.001"),
    taker_percent_fee_decimal=Decimal("0.001"),
    buy_percent_fee_deducted_from_returns=True
)


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    is_spot = False
    is_trading = False

    if exchange_info.get("status", None) == "TRADING":
        is_trading = True

    permissions_sets = exchange_info.get("permissionSets", list())
    for permission_set in permissions_sets:
        # PermissionSet is a list, find if in this list we have "SPOT" value or not
        if "SPOT" in permission_set:
            is_spot = True
            break

    return is_trading and is_spot


class BinanceConfigMap(BaseConnectorConfigMap):
    connector: str = "binance"
    binance_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: "Enter your Binance API key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    binance_api_secret: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": lambda cm: "Enter your Binance API secret",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="binance")


KEYS = BinanceConfigMap.model_construct()



import hashlib
import random
import time
from enum import Enum
from typing import List, Type, TypeVar

import base58


def generate_unique_id():
    timestamp = time.time()
    unique_component = random.randint(0, 99999)
    raw_id = f"{timestamp}-{unique_component}"
    hashed_id = hashlib.sha256(raw_id.encode()).digest()
    return base58.b58encode(hashed_id).decode()


E = TypeVar('E', bound=Enum)


def parse_enum_value(enum_class: Type[E], value, field_name: str = "field") -> E:
    """
    Parse enum from string name or return as-is if already correct type.

    Args:
        enum_class: The enum class to parse into
        value: The value to parse (string name or enum instance)
        field_name: Name of the field for error messages

    Returns:
        The enum value

    Raises:
        ValueError: If the string doesn't match any enum name

    Example:
        >>> from hummingbot.core.data_type.common import TradeType
        >>> parse_enum_value(TradeType, 'BUY', 'side')
        <TradeType.BUY: 1>
    """
    if isinstance(value, str):
        try:
            return enum_class[value.upper()]
        except KeyError:
            valid_names = [e.name for e in enum_class]
            raise ValueError(f"Invalid {field_name}: '{value}'. Expected one of: {valid_names}")
    return value


def parse_comma_separated_list(value) -> List[float]:
    """
    Parse a comma-separated string, scalar number, or list into a List[float].

    Handles values coming from YAML configs where a single value is deserialized
    as a scalar (int/float) rather than a list.

    Args:
        value: The value to parse (str, int, float, list, or None)

    Returns:
        A list of floats, or an empty list if value is None or empty string.

    Example:
        >>> parse_comma_separated_list("0.01,0.02")
        [0.01, 0.02]
        >>> parse_comma_separated_list(0.01)
        [0.01]
    """
    if value is None:
        return []
    if isinstance(value, str):
        if value == "":
            return []
        return [float(x.strip()) for x in value.split(',')]
    if isinstance(value, (int, float)):
        return [float(value)]
    return value



import asyncio
import copy
import logging
import math
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

from hummingbot.core.api_throttler.async_request_context_base import AsyncRequestContextBase
from hummingbot.core.api_throttler.data_types import RateLimit, TaskLog
from hummingbot.logger.logger import HummingbotLogger


class AsyncThrottlerBase(ABC):
    """
    The APIThrottlerBase is an abstract class meant to describe the functions necessary to handle the
    throttling of API requests through the usage of asynchronous context managers.
    """

    _default_config_map = {}
    _logger = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self,
                 rate_limits: List[RateLimit],
                 retry_interval: float = 0.1,
                 safety_margin_pct: Optional[float] = 0.05,  # An extra safety margin, in percentage.
                 limits_share_percentage: Optional[Decimal] = None
                 ):
        """
        :param rate_limits: List of RateLimit(s).
        :param retry_interval: Time between every capacity check.
        :param safety_margin_pct: Percentage of limit to be added as a safety margin when calculating capacity to ensure
            calls are within the limit.
        :param limits_share_percentage: Percentage of the limits to be used by this instance (important when multiple
            bots operate with the same account)
        """
        # If configured, users can define the percentage of rate limits to allocate to the throttler.
        share_percentage = limits_share_percentage or Decimal("100")
        self.limits_pct: Decimal = share_percentage / 100

        self.set_rate_limits(rate_limits)

        # List of TaskLog used to determine the API requests within a set time window.
        self._task_logs: List[TaskLog] = []

        # Throttler Parameters
        self._retry_interval: float = retry_interval
        self._safety_margin_pct: float = safety_margin_pct

        # Shared asyncio.Lock instance to prevent multiple async ContextManager from accessing the _task_logs variable
        self._lock = asyncio.Lock()

    def set_rate_limits(self, rate_limits: List[RateLimit]):
        # Rate Limit Definitions
        self._rate_limits: List[RateLimit] = copy.deepcopy(rate_limits)

        for rate_limit in self._rate_limits:
            rate_limit.limit = max(Decimal("1"), math.floor(Decimal(str(rate_limit.limit)) * self.limits_pct))

        # Dictionary of path_url to RateLimit
        self._id_to_limit_map: Dict[str, RateLimit] = {limit.limit_id: limit for limit in self._rate_limits}

    def add_rate_limits(self, rate_limits: List[RateLimit]):
        """
        Dynamically add new rate limits to the throttler.
        Useful when adding trading pairs at runtime that require pair-specific rate limits.

        :param rate_limits: List of RateLimit(s) to add.
        """
        for rate_limit in rate_limits:
            # Skip if already exists
            if rate_limit.limit_id in self._id_to_limit_map:
                continue
            # Apply the limits percentage
            new_limit = copy.deepcopy(rate_limit)
            new_limit.limit = max(Decimal("1"), math.floor(Decimal(str(new_limit.limit)) * self.limits_pct))
            self._rate_limits.append(new_limit)
            self._id_to_limit_map[new_limit.limit_id] = new_limit

    def _client_config_map(self):
        from hummingbot.client.hummingbot_application import HummingbotApplication  # avoids circular import

        return HummingbotApplication.main_application().client_config_map

    def get_related_limits(self, limit_id: str) -> Tuple[RateLimit, List[Tuple[RateLimit, int]]]:
        rate_limit: Optional[RateLimit] = self._id_to_limit_map.get(limit_id, None)
        linked_limits: List[RateLimit] = [] if rate_limit is None else rate_limit.linked_limits

        related_limits = [(self._id_to_limit_map[limit_weight_pair.limit_id], limit_weight_pair.weight)
                          for limit_weight_pair in linked_limits
                          if limit_weight_pair.limit_id in self._id_to_limit_map]

        # Append self as part of the related_limits
        # if rate_limit is not None:
        #     related_limits.append((rate_limit, rate_limit.weight))
#
        return rate_limit, related_limits

    @abstractmethod
    def execute_task(self, limit_id: str) -> AsyncRequestContextBase:
        raise NotImplementedError



#!/usr/bin/env python

from typing import NamedTuple

from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple


class MakerTakerMarketPair(NamedTuple):
    """
    Specifies a pair of maker and taker markets

    e.g. If I want to market make on DDEX WETH-DAI, and hedge on Binance ETHUSDT... then,
         MakerTakerMarketPair(ddex, "WETH-DAI", "WETH", "DAI",
                          binance, "ETHUSDT", "ETH", "USDT")
    """
    maker: MarketTradingPairTuple
    taker: MarketTradingPairTuple



import time
from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.001"),
    taker_percent_fee_decimal=Decimal("0.001"),
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USDT"


def is_pair_information_valid(pair_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its market information

    :param pair_info: the market information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    return pair_info.get("statusCode") == "Normal"


def get_ms_timestamp() -> int:
    return int(_time() * 1e3)


class AscendExConfigMap(BaseConnectorConfigMap):
    connector: str = "ascend_ex"
    ascend_ex_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your AscendEx API key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ascend_ex_secret_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your AscendEx secret key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ascend_ex_group_id: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your AscendEx group Id",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="ascend_ex")


KEYS = AscendExConfigMap.model_construct()


def _time():
    """
    Private function created just to have a method that can be safely patched during unit tests and make tests
    independent from real time
    """
    return time.time()



import datetime as dt
import glob
import logging
import os
from decimal import Decimal
from typing import List, Tuple

import numpy as np
import pandas as pd

from hummingbot.connector.exchange_base import ExchangeBase

from hummingbot.connector.exchange_base cimport ExchangeBase

from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator

from hummingbot.core.data_type.limit_order cimport LimitOrder

from hummingbot.core.event.events import BuyOrderCompletedEvent, SellOrderCompletedEvent
from hummingbot.strategy.cross_exchange_mining.cross_exchange_mining_config_map_pydantic import (
    CrossExchangeMiningConfigMap,
)
from hummingbot.strategy.strategy_base import StrategyBase

from .cross_exchange_mining_pair import CrossExchangeMiningPair
from .order_id_market_pair_tracker import OrderIDMarketPairTracker

# Cross exchange Mining script by bensmeaton@gmail.com
NaN = float("nan")
s_decimal_zero = Decimal(0)
s_decimal_nan = Decimal("nan")
s_logger = None


cdef class CrossExchangeMiningStrategy(StrategyBase):
    OPTION_LOG_NULL_ORDER_SIZE = 1 << 0
    OPTION_LOG_REMOVING_ORDER = 1 << 1
    OPTION_LOG_ADJUST_ORDER = 1 << 2
    OPTION_LOG_CREATE_ORDER = 1 << 3
    OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
    OPTION_LOG_STATUS_REPORT = 1 << 5
    OPTION_LOG_MAKER_ORDER_HEDGED = 1 << 6
    OPTION_LOG_ALL = 0x7fffffffffffffff

    ORDER_ADJUST_SAMPLE_INTERVAL = 5
    ORDER_ADJUST_SAMPLE_WINDOW = 12

    SHADOW_MAKER_ORDER_KEEP_ALIVE_DURATION = 60.0 * 15
    CANCEL_EXPIRY_DURATION = 60.0

    @classmethod
    def logger(cls):
        global s_logger
        if s_logger is None:
            s_logger = logging.getLogger(__name__)
        return s_logger

    def init_params(self,
                    config_map: CrossExchangeMiningConfigMap,
                    market_pairs: List[CrossExchangeMiningPair],
                    status_report_interval: float = 900,
                    logging_options: int = OPTION_LOG_ALL,
                    hb_app_notification: bool = False,
                    ):
        """
        Initializes a cross exchange market making strategy object.

        :param config_map: Strategy configuration map
        :param market_pairs: list of cross exchange market pairs
        :param logging_options: bit field for what types of logging to enable in this strategy object
        :param hb_app_notification:
        """
        self._config_map = config_map
        self._market_pairs = {
            (market_pair.maker.market, market_pair.maker.trading_pair): market_pair
            for market_pair in market_pairs
        }
        self._maker_markets = set([market_pair.maker.market for market_pair in market_pairs])
        self._taker_markets = set([market_pair.taker.market for market_pair in market_pairs])
        self._all_markets_ready = False
        self._volatility_pct = 0
        self._volatility_timer = 0
        self._balance_timer = 0
        self._maker_side = True
        self._balance_flag = True
        self._tol_o = 0.1 / 100
        self._min_prof_adj = 0
        self._min_prof_adj_t = 0
        self._adjcount = []
        self._avg_vol = InstantVolatilityIndicator(sampling_length=self.volatility_buffer_size)
        self._anti_hysteresis_timers = {}
        self._order_fill_buy_events = {}
        self._order_fill_sell_events = {}
        self._suggested_price_samples = {}
        self._last_timestamp = 0
        self._status_report_interval = status_report_interval
        self._market_pair_tracker = OrderIDMarketPairTracker()
        self._last_conv_rates_logged = 0
        self._hb_app_notification = hb_app_notification

        self._maker_order_ids = []
        cdef:
            list all_markets = list(self._maker_markets | self._taker_markets)

        self.c_add_markets(all_markets)

    @property
    def order_amount(self):
        return self._config_map.order_amount

    @property
    def min_profitability(self):
        return self._config_map.min_profitability / Decimal("100")

    @property
    def avg_vol(self):
        return self._avg_vol

    @avg_vol.setter
    def avg_vol(self, indicator: InstantVolatilityIndicator):
        self._avg_vol = indicator

    @property
    def min_prof_adj_timer(self):
        return self._config_map.min_prof_adj_timer

    @property
    def volatility_buffer_size(self):
        return self._config_map.volatility_buffer_size

    @property
    def min_order_amount(self):
        return self._config_map.min_order_amount

    @property
    def rate_curve(self):
        return self._config_map.rate_curve

    @property
    def trade_fee(self):
        return self._config_map.trade_fee

    @property
    def status_report_interval(self):
        return self._status_report_interval

    @property
    def balance_adjustment_duration(self):
        return self._config_map.balance_adjustment_duration

    @property
    def min_prof_tol_high(self):
        return self._config_map.min_prof_tol_high / Decimal("100")

    @property
    def min_prof_tol_low(self):
        return self._config_map.min_prof_tol_low / Decimal("100")

    @property
    def slippage_buffer(self):
        return self._config_map.slippage_buffer / Decimal("100")

    @property
    def hanging_order_ids(self) -> List[str]:
        return self._hanging_order_ids

    @property
    def active_limit_orders(self) -> List[Tuple[ExchangeBase, LimitOrder]]:
        return [(ex, order) for ex, order in self._sb_order_tracker.active_limit_orders
                if order.client_order_id in self._maker_order_ids]

    def format_status(self) -> str:
        cdef:
            list lines = []
            list warning_lines = []
            dict tracked_maker_orders = {}
            LimitOrder typed_limit_order

        # Go through the currently open limit orders, and group them by market pair.
        active_orders_status = []
        limit_orders = list(self._sb_order_tracker.c_get_limit_orders().values())
        if limit_orders:
            for key in [elem for elem in limit_orders[0]]:
                active_orders_status.append(limit_orders[0][key])

        for market_pair in self._market_pairs.values():
            warning_lines.extend(self.network_warning([market_pair.maker, market_pair.taker]))

            markets_df = self.market_status_data_frame([market_pair.maker, market_pair.taker])
            lines.extend(["", "  Markets:"] +
                         ["    " + line for line in str(markets_df).split("\n")])

            assets_df = self.wallet_balance_data_frame([market_pair.maker, market_pair.taker])
            lines.extend(["", "  Assets:"] +
                         ["    " + line for line in str(assets_df).split("\n")])

            # See if there're any open orders.
            if active_orders_status:
                lines.extend(["", "  Active orders:"])
                for order in active_orders_status:
                    if order.is_buy:
                        lines.extend(["Current buy order of : " + str(order.trading_pair) + " with quantity: " + str(round(order.quantity, 5)) + " of " + str(order.base_currency) + " at price: " + str(round(order.price, 5))])
                    else:
                        lines.extend(["Current sell order of : " + str(order.trading_pair) + " with quantity: " + str(round(order.quantity, 5)) + " of " + str(order.base_currency) + " at price: " + str(round(order.price, 5))])
            else:
                lines.extend(["", "  No active maker orders."])

            warning_lines.extend(self.balance_warning([market_pair.maker, market_pair.taker]))

        if len(warning_lines) > 0:
            lines.extend(["", "  *** WARNINGS ***"] + warning_lines)

        return "\n".join(lines)

    cdef volatility_rate(self, market_pair):
        cdef:
            ExchangeBase maker_market = market_pair.maker.market
            ExchangeBase taker_market = market_pair.taker.market
        mid_price = market_pair.maker.get_mid_price()
        self._avg_vol.add_sample(mid_price)
        vol_abs = Decimal((self._avg_vol.current_value / float(mid_price)) * 3)  # 3 sigma VOLATILITY ADJUSTMENT
        base = Decimal(0.00025)
        if self._volatility_timer + float(self.volatility_buffer_size) < self._current_timestamp or (base * round(vol_abs / base)) >= self._volatility_pct:
            self._volatility_pct = base * round(vol_abs / base)
            self._volatility_timer = self._current_timestamp

    cdef check_order(self, object market_pair, object active_order, object is_buy):
        cdef:
            ExchangeBase taker_market = market_pair.taker.market

        # Check current profitability for existing buy limit order
        if is_buy:
            taker_sell_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, False, active_order.quantity).result_price
            current_buy_price = active_order.price
            current_prof = 1 - (current_buy_price / taker_sell_price)
            # self.notify_hb_app("Check buy order, Current  price: " + str(round(active_order.price,5)) + " Current  qty: " + str(round(active_order.quantity,3)) + " can sell on taker for: " + str(round(taker_sell_price,5)) + " profitability is: " + str(round(current_prof,5)))
        # Check current profitability for existing sell limit order
        if not is_buy:
            taker_buy_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, active_order.quantity).result_price
            current_sell_price = active_order.price
            current_prof = 1 - (taker_buy_price / current_sell_price)
            # self.notify_hb_app("Check sell order, Current  price: " + str(round(active_order.price,5)) + " Current  qty: " + str(round(active_order.quantity,3)) + " can buy on taker for: " + str(round(taker_buy_price,5)) + " profitability is: " + str(round(current_prof,5)))
        # Check profitability is within tolerance around (Min profitability + volatility modifier + long term trade performance modifier)
        # If not cancel limit order
        prof_set = (self.min_profitability + self._volatility_pct + self._min_prof_adj)
        if current_prof < (prof_set - self.min_prof_tol_low) or current_prof > (prof_set + self.min_prof_tol_high):
            # self.notify_hb_app("Cancelling: " + str(current_prof) + " < " + str(prof_set - self.min_prof_tol_low) + " or > "+ str(prof_set + self.min_prof_tol_high))
            # market_trading_pair_tuple = self._sb_order_tracker.c_get_market_pair_from_order_id(active_order.client_order_id)
            market_trading_pair_tuple = self._sb_order_tracker.c_get_market_pair_from_order_id(active_order.client_order_id)
            if market_trading_pair_tuple:
                StrategyBase.c_cancel_order(self, market_trading_pair_tuple, active_order.client_order_id)
                StrategyBase.stop_tracking_limit_order(self, market_trading_pair_tuple, active_order.client_order_id)

    cdef set_order(self, object market_pair, object is_buy):
        cdef:
            ExchangeBase maker_market = market_pair.maker.market
            ExchangeBase taker_market = market_pair.taker.market

        maker_base_side_balance = maker_market.c_get_balance(market_pair.maker.base_asset)
        taker_base_side_balance = taker_market.c_get_balance(market_pair.taker.base_asset)
        try:
            # Average Price for buying order amount on maker market
            maker_buy_price = maker_market.c_get_vwap_for_volume(market_pair.maker.trading_pair, True, self.order_amount).result_price  # True = buy
        except ZeroDivisionError:
            maker_buy_price = taker_market.c_get_vwap_for_volume(market_pair.maker.trading_pair, True, self.order_amount).result_price  # True = buy
        try:
            # Average Price for buying order amount on taker market
            taker_buy_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, self.order_amount).result_price  # True = buy
        except ZeroDivisionError:
            taker_buy_price = maker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, self.order_amount).result_price  # True = buy
        # quantity of maker quote amount in base if sold on maker market
        maker_quote_side_balance_in_base = maker_market.c_get_available_balance(market_pair.maker.quote_asset) / maker_buy_price
        # quantity of taker quote amount in base if sold on taker market
        taker_quote_side_balance_in_base = taker_market.c_get_available_balance(market_pair.taker.quote_asset) / taker_buy_price

        var_list = [maker_base_side_balance, taker_base_side_balance, maker_buy_price, taker_buy_price, maker_quote_side_balance_in_base, taker_quote_side_balance_in_base]
        # self.notify_hb_app("Maker Buffer Price Sell Set: " + str(var_list))
        for item in var_list:
            if Decimal.is_nan(item):
                return s_decimal_nan, s_decimal_nan
        try:
            if not is_buy:  # Sell base on Maker side
                # Sell base on Maker side, take minimum of order amount in base, base amount on maker side and quote amount in base on taker side
                order_amount_maker_sell = min(float(taker_quote_side_balance_in_base), float(maker_base_side_balance), float(self.order_amount))
                if Decimal(order_amount_maker_sell) < self.min_order_amount:
                    return s_decimal_nan, s_decimal_nan
                # you are selling base asset on maker, using the actual order amount calculate the average price you would get if you bought that amount back on the taker side.
                taker_buy_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, Decimal(order_amount_maker_sell)).result_price
                # Therefore you can sell the amount for the price you would get if you bought on the taker market + min profitabilty
                maker_set_price_sell = taker_buy_price * (1 + (self.min_profitability + self._volatility_pct + self._min_prof_adj))
                # self.notify_hb_app("Looking to sell on maker, Can buy " + str(round(order_amount_maker_sell,2)) + " on taker for: " + str(round(taker_buy_price,5)) + " so sell on maker for: " + str(round(maker_set_price_sell,5)))
                return Decimal(maker_set_price_sell), Decimal(order_amount_maker_sell)

            if is_buy:  # buy base on Maker side
                # Buy base on Maker side
                # Buy base on Maker side, take minimum of order amount in base, quote amount on maker side in base and base amount on taker side
                order_amount_maker_buy = min(float(taker_base_side_balance), float(maker_quote_side_balance_in_base), float(self.order_amount))
                if Decimal(order_amount_maker_buy) < self.min_order_amount:
                    return s_decimal_nan, s_decimal_nan
                # you are buying base asset on maker, using the actual order amount calculate the average price you would get if you sold that amount back on the taker side.
                taker_sell_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, False, Decimal(order_amount_maker_buy)).result_price
                # Therefore you can sell the amount for the price you would get if you sold on the taker market - min profitabilty
                maker_set_price_buy = taker_sell_price * (1 - (self.min_profitability + self._volatility_pct + self._min_prof_adj))
                # self.notify_hb_app("Looking to buy on maker, Can sell " + str(round(order_amount_maker_buy,2)) + " on taker for: " + str(round(taker_sell_price,5)) + " so buy on maker for: " + str(round(maker_set_price_buy,5)))
                return Decimal(maker_set_price_buy), Decimal(order_amount_maker_buy)
        except Exception:
            return s_decimal_nan, s_decimal_nan

    cdef check_balance(self, object market_pair):
        cdef:
            ExchangeBase maker_market = market_pair.maker.market
            ExchangeBase taker_market = market_pair.taker.market

        # Script aims to continuosely check and reblance base assets on each exchange to maintain order amount across exchanges. This is also used to complete the XEMM trade after a limit order has been completed.
        maker_base_side_balance = maker_market.c_get_balance(market_pair.maker.base_asset)
        taker_base_side_balance = taker_market.c_get_balance(market_pair.taker.base_asset)
        self._balance_flag = False
        # self.notify_hb_app(str(taker_base_side_balance) +" " + str (maker_base_side_balance) + " " + str(self.order_amount))
        if (self.order_amount - self.min_order_amount) <= (taker_base_side_balance + maker_base_side_balance) <= (self.order_amount + self.min_order_amount):
            return

        try:
            maker_buy_price = maker_market.c_get_vwap_for_volume(market_pair.maker.trading_pair, True, self.order_amount).result_price  # True = buy
        except ZeroDivisionError:
            maker_buy_price = taker_market.c_get_vwap_for_volume(market_pair.maker.trading_pair, True, self.order_amount).result_price  # True = buy
        try:
            taker_buy_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, self.order_amount).result_price  # True = buy
        except ZeroDivisionError:
            taker_buy_price = maker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, self.order_amount).result_price  # True = buy

        maker_quote_side_balance_in_base = maker_market.c_get_available_balance(market_pair.maker.quote_asset) / maker_buy_price
        taker_quote_side_balance_in_base = taker_market.c_get_available_balance(market_pair.taker.quote_asset) / taker_buy_price

        var_list = [maker_base_side_balance, taker_base_side_balance, maker_buy_price, taker_buy_price, maker_quote_side_balance_in_base, taker_quote_side_balance_in_base]

        for item in var_list:
            if Decimal.is_nan(item):
                return

        check_mat = []
        buytaker = False
        selltaker = False

        if taker_base_side_balance + maker_base_side_balance < (self.order_amount - self.min_order_amount):  # Need to Buy taker base balance as maker side does not balance
            taker_qty = Decimal.min((self.order_amount - (taker_base_side_balance + maker_base_side_balance)), taker_quote_side_balance_in_base)
            if taker_qty:
                taker_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, True, Decimal(taker_qty)).result_price
                # self.notify_hb_app(str(taker_qty) + " t " + str(taker_price))
                if taker_qty > self.min_order_amount:
                    buytaker = True
                    check_mat = [False, market_pair, True, taker_qty, taker_price]

        if taker_base_side_balance + maker_base_side_balance > (self.order_amount + self.min_order_amount):  # Need to Sell taker base balance as maker side does not balance
            taker_qty = Decimal.min(((taker_base_side_balance + maker_base_side_balance) - self.order_amount), taker_base_side_balance)
            if taker_qty:
                taker_price = taker_market.c_get_vwap_for_volume(market_pair.taker.trading_pair, False, Decimal(taker_qty)).result_price
                if taker_qty > self.min_order_amount:
                    selltaker = True
                    check_mat = [False, market_pair, False, taker_qty, taker_price]

        if maker_base_side_balance + taker_base_side_balance < (self.order_amount - self.min_order_amount):  # Need to Buy maker base balance as taker side does not balance
            maker_qty = Decimal.min((self.order_amount - (maker_base_side_balance + taker_base_side_balance)), maker_quote_side_balance_in_base)
            if maker_qty:
                maker_price = maker_market.c_get_vwap_for_volume(market_pair.maker.trading_pair, True, Decimal(maker_qty)).result_price
                # self.notify_hb_app(str(maker_qty) + " m " + str(maker_price))
                if maker_qty > self.min_order_amount:
                    if ((maker_price < taker_price and buytaker) or not buytaker):
                        check_mat = [True, market_pair, True, maker_qty, maker_price]

        if maker_base_side_balance + taker_base_side_balance > (self.order_amount + self.min_order_amount):  # Need to Sell maker base balance as taker side does not balance
            maker_qty = Decimal.min(((maker_base_side_balance + taker_base_side_balance) - self.order_amount), maker_base_side_balance)
            if maker_qty:
                maker_price = maker_market.c_get_vwap_for_volume(market_pair.maker.trading_pair, False, Decimal(maker_qty)).result_price
                if maker_qty > self.min_order_amount:
                    if ((maker_price > taker_price and selltaker) or not selltaker):
                        check_mat = [True, market_pair, False, maker_qty, maker_price]
        # self.notify_hb_app(str(check_mat))
        if check_mat:
            self._balance_timer = self._current_timestamp + self.balance_adjustment_duration
            self._balance_flag = True
            return check_mat
        else:
            return

    cdef c_tick(self, double timestamp):
        """
        Clock tick entry point.

        :param timestamp: current tick timestamp
        """
        StrategyBase.c_tick(self, timestamp)
        # Perform clock tick with the market pair tracker.
        self._market_pair_tracker.c_tick(timestamp)

        cdef:
            list active_limit_orders = self.active_limit_orders
            LimitOrder limit_order
            bint has_active_bid = False
            bint has_active_ask = False

        if not self._all_markets_ready:
            self._all_markets_ready = all([market.ready for market in self._sb_markets])
            if not self._all_markets_ready:
                # Markets not ready yet. Don't do anything.
                return
            else:
                # Markets are ready, ok to proceed.
                if self.OPTION_LOG_STATUS_REPORT:
                    self.logger().info("Markets are ready. Trading started.")

        for market_pair in self._market_pairs.values():
            buy_orders = []
            sell_orders = []
            active_orders = []

            # Check for active limit orders and create buy and sell order lists
            limit_orders = list(self._sb_order_tracker.c_get_limit_orders().values())
            if limit_orders:
                for key in [elem for elem in limit_orders[0]]:
                    active_orders.append(limit_orders[0][key])
                    if limit_orders[0][key].is_buy:
                        buy_orders.append(limit_orders[0][key])
                    else:
                        sell_orders.append(limit_orders[0][key])

            self.volatility_rate(market_pair)
            # If there are buy orders check buy orders
            if buy_orders:
                for active_buy_order in buy_orders:
                    self.check_order(market_pair, active_buy_order, True)

            # If there are sell orders check sell orders
            if sell_orders:
                for active_sell_order in sell_orders:
                    self.check_order(market_pair, active_sell_order, False)

            # If there are no buy orders set them
            if not buy_orders:
                buyprice, buysize = self.set_order(market_pair, True)
                if not Decimal.is_nan(buyprice):
                    if not self._balance_flag:
                        self.c_place_order(market_pair, True, True, buysize, buyprice, True)

            # If there are no sell orders set them
            if not sell_orders:
                sellprice, sellsize = self.set_order(market_pair, False)
                if not Decimal.is_nan(sellprice):
                    if not self._balance_flag:
                        self.c_place_order(market_pair, False, True, sellsize, sellprice, True)

            # Balance base across exchanges
            if self._current_timestamp > self._balance_timer:
                bal_check = self.check_balance(market_pair)
                if bal_check:
                    # clear any existing trades to prioritise balance
                    self.cancel_all_trades(active_orders)
                    if bal_check[2]:  # buy order
                        self.c_place_order(bal_check[1], True, bal_check[0], bal_check[3], bal_check[4] * (1 + self.slippage_buffer), False)
                        return
                    else:  # sell order
                        self.c_place_order(bal_check[1], False, bal_check[0], bal_check[3], bal_check[4] * (1 - self.slippage_buffer), False)
                        return

            if self._current_timestamp > self._min_prof_adj_t:
                self._min_prof_adj = Decimal(self.adjust_profitability_lag(market_pair))
                self._min_prof_adj_t = self._current_timestamp + self.min_prof_adj_timer
                self.cancel_all_trades(active_orders)
                self.logger().info("Adjusted Min Profitability from previous trades")

    cdef adjust_profitability_lag(self, object market_pair):
        cdef:
            ExchangeBase maker_market = market_pair.maker.market
            ExchangeBase taker_market = market_pair.taker.market

        feea1 = float(self.trade_fee)
        min_prof = 0
        rate_curve = float(self.rate_curve)
        try:
            fileloc = max(glob.glob('data/trades_*.csv'), key=os.path.getmtime)
        except Exception:
            return 0
        df = pd.read_csv(fileloc)
        df["timestamp"] = pd.to_datetime(df["timestamp"], unit='ms')
        # Show last 24 hour results
        df = df[df['timestamp'] >= (dt.datetime.now() - dt.timedelta(hours=24))].reset_index(drop=True)
        an_1 = []
        while (len(df) > 0):
            try:
                start = df['timestamp'][0] - pd.Timedelta(seconds=1)
                stop = df['timestamp'][0] + pd.Timedelta(minutes=1)
                mattime = df.loc[(df['timestamp'] > start) & (df['timestamp'] < stop)].reindex(columns = ['order_id', 'symbol', 'price', 'amount', 'timestamp', 'trade_type']).reset_index(drop=True)
                mattime = mattime.loc[(mattime['price'] < mattime['price'][0] * 1.1) & (mattime['price'] > mattime['price'][0] * 0.9)]
                mattime = mattime.loc[(mattime['trade_type'] == mattime['trade_type'][0])].reset_index(drop=True)
                if mattime['trade_type'][0] == 'BUY':
                    sidex = 'SELL'
                else:
                    sidex = 'BUY'
                tattime = df.loc[(df['timestamp'] > start) & (df['timestamp'] < stop)].reindex(columns = ['order_id', 'symbol', 'price', 'amount', 'timestamp', 'trade_type']).reset_index(drop=True)
                tattime = tattime.loc[(tattime['price'] < mattime['price'][0] * 1.1) & (tattime['price'] > mattime['price'][0] * 0.9)]
                tattime = tattime.loc[(tattime['trade_type'] == sidex)].reset_index(drop=True)
                df = df.iloc[len(mattime):].reset_index(drop=True)
                df = df.iloc[len(tattime):].reset_index(drop=True)
                s_id = str(tattime['order_id'][0]) + str(mattime['order_id'][0])
                s_msum = mattime['amount'].astype(float).sum()
                s_tsum = tattime['amount'].astype(float).sum()
                s_mside = mattime['trade_type'][0]
                s_tside = tattime['trade_type'][0]
                if s_msum < s_tsum * 0.8 or s_msum > s_tsum * 1.2:
                    tattime = (tattime.iloc[(tattime['amount'] - s_msum).abs().argsort()[:1]]).reset_index(drop=True)
                    s_tsum = tattime['amount'].astype(float).sum()
                s_pmak = 0
                for x in range(len(mattime)):
                    s_pmak += float(mattime['price'][x]) * float(mattime['amount'][x])
                s_pmak = s_pmak / s_msum
                s_ptak = 0
                for y in range(len(tattime)):
                    s_ptak += float(tattime['price'][y]) * float(tattime['amount'][y])
                s_ptak = s_ptak / s_tsum
                sym_q = mattime['symbol'][0].split('-')[1]
                if mattime['trade_type'][0] == 'BUY':
                    s_per = ((s_ptak - s_pmak) / ((s_ptak + s_pmak) / 2)) * 100
                if mattime['trade_type'][0] == 'SELL':
                    s_per = ((s_pmak - s_ptak) / ((s_ptak + s_pmak) / 2)) * 100
                base = ((s_ptak + s_pmak) / 2) * ((s_msum + s_tsum) / 2) * ((s_per - feea1) / 100)
                per_am = (s_per - feea1) * ((s_msum + s_tsum) / 2)
                per_am_sum = ((s_msum + s_tsum) / 2)
                an_1.append([s_id, (tattime['timestamp'][0] - mattime['timestamp'][0]) / np.timedelta64(1, 's'), mattime['symbol'][0], s_msum, s_tsum, s_mside, s_tside, s_pmak, s_ptak, s_per, s_per - feea1, base, per_am, per_am_sum])
            except Exception:
                pass
        an_final = pd.DataFrame(an_1, columns = ['id', 'time', 'symbol', 'Total_Maker', 'Total_Taker', 'Maker_Side', 'Taker_Side', 'Price_Maker', 'Price_Taker', 'Percentage', 'Percentage_+_Fee', 'cost (quote)', 'per_am', 'per_am_sum'])
        if len(an_final) == 0:
            self.logger().info("No trade records within the past 24 hours found within datafile")
            return 0
        per_am = an_final['per_am'].sum() / an_final['per_am_sum'].sum()
        sum_base = an_final['cost (quote)'].sum()
        time_av = an_final['time'].mean()
        prof_adj = (- (per_am * rate_curve)**3 + min_prof) / 100
        self.logger().info(f"Percentage for trades over the last 24 Hours: {round(per_am, 3)}, No trades: {round(len(an_final))}, Min Profitability: {round(self.min_profitability + Decimal(prof_adj), 5)}, Profit: {round(sum_base, 2)} {sym_q}, Mean time to balance after fill: {round(time_av, 2)} s")
        return prof_adj

    cdef str c_place_order(self, object market_pair, bint is_buy, bint is_maker, object amount, object price, bint is_limit):

        cdef:
            str order_id
            object market_info = market_pair.maker if is_maker else market_pair.taker
            object order_type = market_info.market.get_maker_order_type() if is_limit else market_info.market.get_taker_order_type()

        amount = amount * Decimal(1 - self._tol_o)
        if is_buy:
            order_id = StrategyBase.c_buy_with_specific_market(self, market_info, amount, order_type=order_type, price=price, expiration_seconds=NaN)
        else:
            order_id = StrategyBase.c_sell_with_specific_market(self, market_info, amount, order_type=order_type, price=price, expiration_seconds=NaN)
        # market_trading_pair_tuple = self._sb_order_tracker.c_get_market_pair_from_order_id(order_id)
        mid_price = market_info.get_mid_price()
        if is_limit:
            StrategyBase.start_tracking_limit_order(self, market_info, order_id, is_buy, price, amount)
            self.logger().info(f"Maker Order Created: {market_pair.maker.trading_pair}, Amount: {round(amount, 3)}, Price: {round(price, 3)}, Min profitability: {self.min_profitability} + {round(self._volatility_pct, 5)} + {round(self._min_prof_adj, 5)}, % from Mid Market: {round(((max(price, mid_price)/min(price, mid_price))-1)*100, 3)} ")
        else:
            self.logger().info(f"Taker Order Created: {market_pair.maker.trading_pair}, Amount: {round(amount, 3)}, Price: {round(price, 3)}, Min profitability: {self.min_profitability} + {round(self._volatility_pct, 5)} + {round(self._min_prof_adj, 5)}, % from Mid Market: {round(((max(price, mid_price)/min(price, mid_price))-1)*100, 3)} ")
            # self.notify_hb_app("Order Created: " + str(market_pair.maker.trading_pair) +" Amount: " + str(amount) + " Price: " + str(price) + " Min profitability: " + str(self.min_profitability) + " + " + str(self._volatility_pct) + " + " + str(self._min_prof_adj))
        return order_id

    def notify_hb_app(self, msg: str):
        if self._hb_app_notification:
            super().notify_hb_app(msg)

    def cancel_all_trades(self, active_orders):
        for active_order in active_orders:
            market_trading_pair_tuple = self._sb_order_tracker.c_get_market_pair_from_order_id(active_order.client_order_id)
            if market_trading_pair_tuple:
                StrategyBase.c_cancel_order(self, market_trading_pair_tuple, active_order.client_order_id)
                StrategyBase.stop_tracking_limit_order(self, market_trading_pair_tuple, active_order.client_order_id)

    def did_complete_buy_order(self, order_completed_event: BuyOrderCompletedEvent):
        """
        Output log message when a bid order (on maker side or taker side) is completely taken.
        :param order_completed_event: event object
        """
        order_id = order_completed_event.order_id
        market_pair = self._market_pair_tracker.get_market_pair_from_order_id(order_id)
        if market_pair is not None:
            limit_order_record = self._sb_order_tracker.get_limit_order(market_pair.maker, order_id)
            self.notify_hb_app("Buy Order Completed: " + str(market_pair.maker.trading_pair) + " Quantity: " + str(limit_order_record.quantity) + " of " + str(limit_order_record.base_currency) + " Price: " + str(limit_order_record.price))

    def did_complete_sell_order(self, order_completed_event: SellOrderCompletedEvent):
        """
        Output log message when a ask order (on maker side or taker side) is completely taken.
        :param order_completed_event: event object
        """
        order_id = order_completed_event.order_id
        market_pair = self._market_pair_tracker.get_market_pair_from_order_id(order_id)

        if market_pair is not None:
            limit_order_record = self._sb_order_tracker.get_limit_order(market_pair.maker, order_id)
            self.notify_hb_app("Sell Order Completed: " + str(market_pair.maker.trading_pair) + " Quantity: " + str(limit_order_record.quantity) + " of " + str(limit_order_record.base_currency) + " Price: " + str(limit_order_record.price))



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PublicAggreBookTickerV3Api(_message.Message):
    __slots__ = ("bidPrice", "bidQuantity", "askPrice", "askQuantity")
    BIDPRICE_FIELD_NUMBER: _ClassVar[int]
    BIDQUANTITY_FIELD_NUMBER: _ClassVar[int]
    ASKPRICE_FIELD_NUMBER: _ClassVar[int]
    ASKQUANTITY_FIELD_NUMBER: _ClassVar[int]
    bidPrice: str
    bidQuantity: str
    askPrice: str
    askQuantity: str
    def __init__(self, bidPrice: _Optional[str] = ..., bidQuantity: _Optional[str] = ..., askPrice: _Optional[str] = ..., askQuantity: _Optional[str] = ...) -> None: ...



from typing import List, Optional

from hummingbot.core.api_throttler.async_throttler_base import AsyncThrottlerBase
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.connections_factory import ConnectionsFactory
from hummingbot.core.web_assistant.rest_assistant import RESTAssistant
from hummingbot.core.web_assistant.rest_post_processors import RESTPostProcessorBase
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.core.web_assistant.ws_post_processors import WSPostProcessorBase
from hummingbot.core.web_assistant.ws_pre_processors import WSPreProcessorBase


class WebAssistantsFactory:
    """Creates `RESTAssistant` and `WSAssistant` objects.

    The purpose of the `web_assistant` layer is to abstract away all WebSocket and REST operations from the exchange
    logic. The assistant objects are designed to be injectable with additional logic via the pre- and post-processor
    lists. Consult the documentation of the relevant assistant and/or pre-/post-processor class for
    additional information.

    todo: integrate AsyncThrottler
    """

    def __init__(
        self,
        throttler: AsyncThrottlerBase,
        rest_pre_processors: Optional[List[RESTPreProcessorBase]] = None,
        rest_post_processors: Optional[List[RESTPostProcessorBase]] = None,
        ws_pre_processors: Optional[List[WSPreProcessorBase]] = None,
        ws_post_processors: Optional[List[WSPostProcessorBase]] = None,
        auth: Optional[AuthBase] = None,
        connections_factory: Optional[ConnectionsFactory] = None,
    ):
        self._connections_factory = connections_factory or ConnectionsFactory()
        self._rest_pre_processors = rest_pre_processors or []
        self._rest_post_processors = rest_post_processors or []
        self._ws_pre_processors = ws_pre_processors or []
        self._ws_post_processors = ws_post_processors or []
        self._auth = auth
        self._throttler = throttler

    @property
    def throttler(self) -> AsyncThrottlerBase:
        return self._throttler

    @property
    def auth(self) -> Optional[AuthBase]:
        return self._auth

    async def get_rest_assistant(self) -> RESTAssistant:
        connection = await self._connections_factory.get_rest_connection()
        assistant = RESTAssistant(
            connection=connection,
            throttler=self._throttler,
            rest_pre_processors=self._rest_pre_processors,
            rest_post_processors=self._rest_post_processors,
            auth=self._auth
        )
        return assistant

    async def get_ws_assistant(self) -> WSAssistant:
        connection = await self._connections_factory.get_ws_connection()
        assistant = WSAssistant(
            connection, self._ws_pre_processors, self._ws_post_processors, self._auth
        )
        return assistant

    async def close(self) -> None:
        """
        Close the underlying connections.
        """
        await self._connections_factory.close()

    async def __aenter__(self) -> "WebAssistantsFactory":
        # If the underlying connections factory is a context manager, enter its context.
        if hasattr(self._connections_factory, "__aenter__"):
            await self._connections_factory.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()



from typing import Union

from prompt_toolkit.styles import Style
from prompt_toolkit.utils import is_windows

from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import ClientConfigAdapter, save_to_yml
from hummingbot.client.settings import CLIENT_CONFIG_PATH


def load_style(config_map: ClientConfigAdapter):
    """
    Return a dict mapping {ui_style_name -> style_dict}.
    """
    config_map: Union[ClientConfigAdapter, ClientConfigMap] = config_map  # to enable IDE auto-complete
    # Load config
    color_top_pane = config_map.color.top_pane
    color_bottom_pane = config_map.color.bottom_pane
    color_output_pane = config_map.color.output_pane
    color_input_pane = config_map.color.input_pane
    color_logs_pane = config_map.color.logs_pane
    color_terminal_primary = config_map.color.terminal_primary

    color_primary_label = config_map.color.primary_label
    color_secondary_label = config_map.color.secondary_label
    color_success_label = config_map.color.success_label
    color_warning_label = config_map.color.warning_label
    color_info_label = config_map.color.info_label
    color_error_label = config_map.color.error_label
    color_gold_label = config_map.color.gold_label
    color_silver_label = config_map.color.silver_label
    color_bronze_label = config_map.color.bronze_label

    if is_windows():
        # Load default style for Windows
        style = win32_code_style

        # Translate HEX to ANSI
        color_top_pane = hex_to_ansi(color_top_pane)
        color_bottom_pane = hex_to_ansi(color_bottom_pane)
        color_output_pane = hex_to_ansi(color_output_pane)
        color_input_pane = hex_to_ansi(color_input_pane)
        color_logs_pane = hex_to_ansi(color_logs_pane)
        color_terminal_primary = hex_to_ansi(color_terminal_primary)

        color_primary_label = hex_to_ansi(color_primary_label)
        color_secondary_label = hex_to_ansi(color_secondary_label)
        color_success_label = hex_to_ansi(color_success_label)
        color_warning_label = hex_to_ansi(color_warning_label)
        color_info_label = hex_to_ansi(color_info_label)
        color_error_label = hex_to_ansi(color_error_label)
        color_gold_label = hex_to_ansi(color_gold_label)
        color_silver_label = hex_to_ansi(color_silver_label)
        color_bronze_label = hex_to_ansi(color_bronze_label)

        # Apply custom configuration
        style["output_field"] = "bg:" + color_output_pane + " " + color_terminal_primary
        style["input_field"] = "bg:" + color_input_pane + " " + style["input_field"].split(' ')[-1]
        style["log_field"] = "bg:" + color_logs_pane + " " + style["log_field"].split(' ')[-1]
        style["tab_button.focused"] = "bg:" + color_terminal_primary + " " + color_logs_pane
        style["tab_button"] = style["tab_button"].split(' ')[0] + " " + color_logs_pane
        style["header"] = "bg:" + color_top_pane + " " + style["header"].split(' ')[-1]
        style["footer"] = "bg:" + color_bottom_pane + " " + style["footer"].split(' ')[-1]
        style["primary"] = color_terminal_primary
        style["dialog.body"] = style["dialog.body"].split(' ')[0] + " " + color_terminal_primary
        style["dialog frame.label"] = "bg:" + color_terminal_primary + " " + style["dialog frame.label"].split(' ')[-1]
        style["text-area"] = style["text-area"].split(' ')[0] + " " + color_terminal_primary
        style["search"] = color_terminal_primary
        style["search.current"] = color_terminal_primary

        style["primary_label"] = "bg:" + color_primary_label + " " + color_output_pane
        style["secondary_label"] = "bg:" + color_secondary_label + " " + color_output_pane
        style["success_label"] = "bg:" + color_success_label + " " + color_output_pane
        style["warning_label"] = "bg:" + color_warning_label + " " + color_output_pane
        style["info_label"] = "bg:" + color_info_label + " " + color_output_pane
        style["error_label"] = "bg:" + color_error_label + " " + color_output_pane
        style["gold_label"] = "bg:" + color_output_pane + " " + color_gold_label
        style["silver_label"] = "bg:" + color_output_pane + " " + color_silver_label
        style["bronze_label"] = "bg:" + color_output_pane + " " + color_bronze_label

        return Style.from_dict(style)

    else:
        # Load default style
        style = default_ui_style

        # Apply custom configuration
        style["output_field"] = "bg:" + color_output_pane + " " + color_terminal_primary
        style["input_field"] = "bg:" + color_input_pane + " " + style["input_field"].split(' ')[-1]
        style["log_field"] = "bg:" + color_logs_pane + " " + style["log_field"].split(' ')[-1]
        style["header"] = "bg:" + color_top_pane + " " + style["header"].split(' ')[-1]
        style["footer"] = "bg:" + color_bottom_pane + " " + style["footer"].split(' ')[-1]
        style["primary"] = color_terminal_primary
        style["dialog.body"] = style["dialog.body"].split(' ')[0] + " " + color_terminal_primary
        style["dialog frame.label"] = "bg:" + color_terminal_primary + " " + style["dialog frame.label"].split(' ')[-1]
        style["text-area"] = style["text-area"].split(' ')[0] + " " + color_terminal_primary
        style["tab_button.focused"] = "bg:" + color_terminal_primary + " " + color_logs_pane
        style["tab_button"] = style["tab_button"].split(' ')[0] + " " + color_logs_pane

        style["primary_label"] = "bg:" + color_primary_label + " " + color_output_pane
        style["secondary_label"] = "bg:" + color_secondary_label + " " + color_output_pane
        style["success_label"] = "bg:" + color_success_label + " " + color_output_pane
        style["warning_label"] = "bg:" + color_warning_label + " " + color_output_pane
        style["info_label"] = "bg:" + color_info_label + " " + color_output_pane
        style["error_label"] = "bg:" + color_error_label + " " + color_output_pane
        style["gold_label"] = "bg:" + color_output_pane + " " + color_gold_label
        style["silver_label"] = "bg:" + color_output_pane + " " + color_silver_label
        style["bronze_label"] = "bg:" + color_output_pane + " " + color_bronze_label
        return Style.from_dict(style)


def reset_style(config_map: ClientConfigAdapter, save=True):
    # Reset config

    config_map.color.top_pane = config_map.color.get_default("top_pane")
    config_map.color.bottom_pane = config_map.color.get_default("bottom_pane")
    config_map.color.output_pane = config_map.color.get_default("output_pane")
    config_map.color.input_pane = config_map.color.get_default("input_pane")
    config_map.color.logs_pane = config_map.color.get_default("logs_pane")
    config_map.color.terminal_primary = config_map.color.get_default("terminal_primary")

    config_map.color.primary_label = config_map.color.get_default("primary_label")
    config_map.color.secondary_label = config_map.color.get_default("secondary_label")
    config_map.color.success_label = config_map.color.get_default("success_label")
    config_map.color.warning_label = config_map.color.get_default("warning_label")
    config_map.color.info_label = config_map.color.get_default("info_label")
    config_map.color.error_label = config_map.color.get_default("error_label")
    config_map.color.gold_label = config_map.color.get_default("gold_label")
    config_map.color.silver_label = config_map.color.get_default("silver_label")
    config_map.color.bronze_label = config_map.color.get_default("bronze_label")

    # Save configuration
    if save:
        save_to_yml(CLIENT_CONFIG_PATH, config_map)

    # Apply & return style
    return load_style(config_map)


def hex_to_ansi(color_hex):
    ansi_palette = {"000000": "ansiblack",
                    "FF0000": "ansired",
                    "00FF00": "ansigreen",
                    "FFFF00": "ansiyellow",
                    "0000FF": "ansiblue",
                    "FF00FF": "ansimagenta",
                    "00FFFF": "ansicyan",
                    "F0F0F0": "ansigray",
                    "FFFFFF": "ansiwhite",
                    "FFD700": "ansiyellow",
                    "C0C0C0": "ansilightgray",
                    "CD7F32": "ansibrown"
                    }

    # Sanitization
    color_hex = color_hex.replace('#', '')

    # Calculate distance, choose the closest ANSI color
    hex_r = int(color_hex[0:2], 16)
    hex_g = int(color_hex[2:4], 16)
    hex_b = int(color_hex[4:6], 16)

    distance_min = None

    for ansi_hex in ansi_palette:
        ansi_r = int(ansi_hex[0:2], 16)
        ansi_g = int(ansi_hex[2:4], 16)
        ansi_b = int(ansi_hex[4:6], 16)

        distance = abs(ansi_r - hex_r) + abs(ansi_g - hex_g) + abs(ansi_b - hex_b)

        if distance_min is None or distance < distance_min:
            distance_min = distance
            color_ansi = ansi_palette[ansi_hex]

    return "#" + color_ansi


text_ui_style = {
    "&cGOLD": "gold_label",
    "&cSILVER": "silver_label",
    "&cBRONZE": "bronze_label",
}

default_ui_style = {
    "output_field":               "bg:#171E2B #1CD085",  # noqa: E241
    "input_field":                "bg:#000000 #FFFFFF",  # noqa: E241
    "log_field":                  "bg:#171E2B #FFFFFF",  # noqa: E241
    "header":                     "bg:#000000 #AAAAAA",  # noqa: E241
    "footer":                     "bg:#000000 #AAAAAA",  # noqa: E241
    "search":                     "bg:#000000 #93C36D",  # noqa: E241
    "search.current":             "bg:#000000 #1CD085",  # noqa: E241
    "primary":                    "#1CD085",  # noqa: E241
    "warning":                    "#93C36D",  # noqa: E241
    "error":                      "#F5634A",  # noqa: E241
    "tab_button.focused":         "bg:#1CD085 #171E2B",  # noqa: E241
    "tab_button":                 "bg:#FFFFFF #000000",  # noqa: E241
    "dialog": "bg:#171E2B",
    "dialog frame.label": "bg:#FFFFFF #000000",
    "dialog.body": "bg:#000000 ",
    "dialog shadow": "bg:#171E2B",
    "button": "bg:#FFFFFF #000000",
    "text-area": "bg:#000000 #FFFFFF",
}


# Style for an older version of Windows consoles. They support only 16 colors,
# so we choose a combination that displays nicely.
win32_code_style = {
    "output_field":               "#ansigreen",  # noqa: E241
    "input_field":                "#ansiwhite",  # noqa: E241
    "log_field":                  "#ansiwhite",  # noqa: E241
    "header":                     "#ansiwhite",  # noqa: E241
    "footer":                     "#ansiwhite",  # noqa: E241
    "search":                     "#ansigreen",  # noqa: E241
    "search.current":             "#ansigreen",  # noqa: E241
    "primary":                    "#ansigreen",  # noqa: E241
    "warning":                    "#ansibrightyellow",  # noqa: E241
    "error":                      "#ansired",  # noqa: E241
    "tab_button.focused":         "bg:#ansigreen #ansiblack",  # noqa: E241
    "tab_button":                 "bg:#ansiwhite #ansiblack",  # noqa: E241
    "dialog": "bg:#ansigreen",
    "dialog frame.label": "bg:#ansiwhite #ansiblack",
    "dialog.body": "bg:#ansiblack ",
    "dialog shadow": "bg:#ansigreen",
    "button": "bg:#ansiwhite #ansiblack",
    "text-area": "bg:#ansiblack #ansigreen",
}



from hummingbot.core.event.event_reporter cimport EventReporter
from hummingbot.core.event.event_logger cimport EventLogger
from hummingbot.core.data_type.order_book cimport OrderBook
from hummingbot.connector.connector_base cimport ConnectorBase
from hummingbot.core.data_type.order_book_query_result cimport(
    ClientOrderBookQueryResult,
    OrderBookQueryResult,
)

cdef class ExchangeBase(ConnectorBase):
    cdef:
        object _order_book_tracker
        object _budget_checker
        object _trading_pair_symbol_map
        object _mapping_initialization_lock

    cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
    cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
    cdef c_cancel(self, str trading_pair, str client_order_id)
    cdef c_stop_tracking_order(self, str order_id)
    cdef OrderBook c_get_order_book(self, str trading_pair)
    cdef object c_get_price(self, str trading_pair, bint is_buy)
    cdef ClientOrderBookQueryResult c_get_quote_volume_for_base_amount(
        self,
        str trading_pair,
        bint is_buy,
        object base_amount)
    cdef ClientOrderBookQueryResult c_get_volume_for_price(self, str trading_pair, bint is_buy, object price)
    cdef ClientOrderBookQueryResult c_get_quote_volume_for_price(self, str trading_pair, bint is_buy, object price)
    cdef ClientOrderBookQueryResult c_get_vwap_for_volume(self, str trading_pair, bint is_buy, object volume)
    cdef ClientOrderBookQueryResult c_get_price_for_quote_volume(self, str trading_pair, bint is_buy, double volume)
    cdef ClientOrderBookQueryResult c_get_price_for_volume(self, str trading_pair, bint is_buy, object volume)
    cdef object c_get_fee(
        self,
        str base_currency,
        str quote_currency,
        object order_type,
        object order_side,
        object amount,
        object price,
        object is_maker= *)
